//iriver firmware decoder
//created by Khassanov Ural <ural@uwc.ufanet.ru>, 18.07.2002
//linux compilation: c++ -o idecode d.cpp
//
//help to get this hack work:
//  - get valid riovolt or iriver firmware (imp-XXX.hex file)
//    !!!soul firmwares MUST be de'xor-ed first (soul sucks)!!!
//  - apply this program
//    (if it segfaults, you probably download a bad file)

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <new.h>

typedef unsigned char byte;
typedef unsigned int uint; //32bit is a must

typedef struct
{
  char h_id[16];
  char h_mnf[16];
  uint h_size;
  uint h_checksum;
}IHEADER;

byte *p_cbuf;
byte *p_dbuf;
uint csize;
uint cbyte;
byte cbitno;

typedef struct
{
  byte d0[32];
  uint d32;
  uint d36;
}NODESTRUCT;

typedef struct linkstruct_s
{
  uint d0;
  struct linkstruct_s *d4;
  struct linkstruct_s *d8;
}LINKSTRUCT;

NODESTRUCT nodes[257];
LINKSTRUCT links[100000]; //how much it should be? Buy 1GB RAM:)

//-------------------------------------------------------
byte getbit();
uint getbits(byte n);
void process_tree();
void decode();

//-------------------------------------------------------
int main(int argc, char *argv[])
{
  set_new_handler(0);
  puts("IRiver firmware decoder. Written by Ural Khassanov <ural@uwc.ufanet.ru>.");
  if (argc < 3)
  {
    printf("Hint: %s inputfile outputfile\n", argv[0]);
    exit(EXIT_SUCCESS);
  }
  
  FILE *f;
  f = fopen(argv[1], "rb");
  if (f == NULL) { puts("Can't open input file."); exit(EXIT_FAILURE); }
  fseek(f, 0, SEEK_END);
  long int fs = ftell(f);
  if (fs == -1) { puts("Unknown file size."); exit(EXIT_FAILURE); }
  rewind(f);
  IHEADER hdr;
  if (fread(&hdr, sizeof(hdr), 1, f) != 1) { puts("Read error."); exit(EXIT_FAILURE); }
  hdr.h_id[15] = hdr.h_mnf[15] = 0;
  printf("Header information\nID: %s\nManufacturer: %s\n", hdr.h_id, hdr.h_mnf);
  printf("Size: %u\nChecksum: %u\n", hdr.h_size, hdr.h_checksum);
  fs-=40;
  if (strcmp(hdr.h_id, "dmp-01") == 0 || strcmp(hdr.h_id, "DMP-01") == 0)
  { 
    puts("SOUL firmwares are not supported. Convert them to IRiver format first."); 
    exit(EXIT_FAILURE); 
  }
  if ((uint)fs != hdr.h_size)
  { puts("Header and file sizes differ. Bad file?"); exit(EXIT_FAILURE); }
  csize = (fs + 511) & 0xFFFFFE00;
  byte *p_c = new byte[csize];
  byte *p_d = new byte[0x80000*4];
  if (p_c == NULL || p_d == NULL)
  { puts("Memory allocation error."); exit(EXIT_FAILURE); }
  p_cbuf = p_c; memset(p_cbuf, 0, csize);
  p_dbuf = p_d; memset(p_dbuf, 0, 0x80000);
  if (fread(p_c, fs, 1, f) != 1) { puts("Read error."); exit(EXIT_FAILURE); }
  fclose(f);
  
  cbitno = 0;
  cbyte = 0;

  if (csize > 0)
  {
    process_tree();
    decode();
    
    uint d;
    LINKSTRUCT *plink;
    do
    {
      plink = links;
      while (plink->d0 == 257/*ESCAPE*/)
      {
        if (getbit() == 0)
     	  plink = plink->d8;
        else
	  plink = plink->d4;
      }
      d = plink->d0;
      if (d <= 255)
      {
        *p_dbuf = (byte)d;
        p_dbuf++;
      }
    }while(d != 256/*EOFSTREAM*/);
  }//if (csize > 0)
  
  uint dfs = p_dbuf - p_d;
  printf("Block of %u bytes decoded to %u bytes.\n", (uint)fs, dfs);

  f = fopen(argv[2], "wb");
  if (f == NULL) { puts("Can't create output file."); exit(EXIT_FAILURE); }
  if (dfs > 0 && fwrite(p_d, dfs, 1, f) != 1)
  { puts("Write error."); exit(EXIT_FAILURE); }
  fclose(f);
  
  delete[] p_d;
  delete[] p_c;
  return 0;
}
/*main()*/

//-------------------------------------------------------
byte getbit()
{
  uint r;
  if (cbitno == 0)
  {
    cbyte = *p_cbuf;
    p_cbuf++;
    cbitno = 7;
    r = (cbyte >> 7) & 1;
  }
  else
  {
    cbitno--;
    r = (cbyte >> cbitno) & 1;
  }
  cbyte &= 127;
  return r;
}
/*getbit()*/

uint getbits(byte n)
{
  uint r;
  while (cbitno < n)
  {
    cbyte = (cbyte << 8) + *p_cbuf;
    p_cbuf++;
    cbitno += 8;
  }
  cbitno -= n;
  r = cbyte >> cbitno;
  cbyte &= (1 << cbitno) - 1;
  return r;
}
/*getbits()*/

//-------------------------------------------------------
void process_tree()
{
  uint cnt, offs, d;
  memset(nodes, 0, sizeof(nodes));
  if (getbit() != 0)
  {
    cnt = 0;
    while (cnt <= 255)
    {
      nodes[cnt].d36 = getbit();
      cnt++;
    }
  }
  else
  {
    cnt = getbits(5) + 1;
    while (cnt != 0)
    {
      nodes[getbits(8)].d36 = 1;
      cnt--;
    }
  }

  nodes[256].d36 = 1;
  cnt = 0;
  while (cnt <= 256)
  {
    if (nodes[cnt].d36 != 0)
    {
      if (getbit() != 0)
        d = getbits(8) + 1;
      else
	d = getbits(5) + 1;
      nodes[cnt].d32 = d;
      offs = ((d + 7) >> 3) & 0xFF;
      if ((d & 7) != 0)
      {
	offs = (offs - 1) & 0xFF;
        *(nodes[cnt].d0 + offs) = (byte)getbits(d & 7);
      }
      while (offs != 0)
      {
	offs = (offs - 1) & 0xFF;
        *(nodes[cnt].d0 + offs) = (byte)getbits(8);
      }
    }
    cnt++;
  }
}
/*process_tree()*/

//-------------------------------------------------------
void decode()
{
  uint t, cnt, d, loffs, dn;
  LINKSTRUCT *plink;
  links[0].d0 = 257;
  links[0].d4 = NULL;
  links[0].d8 = NULL;

  loffs = 1;
  cnt = 0;
  while (cnt <= 256)
  {
    plink = links;
    dn = nodes[cnt].d32;
    while (dn != 0)
    {
      t = dn - 1;
      d = *(nodes[cnt].d0 + (t >> 3));
      if ((d & (1 << (t & 7))) != 0)
      {
	if (plink->d4 == NULL)
	{
          plink->d4 = links + loffs;
	  plink = plink->d4;
	  plink->d4 = NULL;
	  plink->d8 = NULL;
	  loffs++;
	}
	else
	  plink = plink->d4;
      }
      else
      {
	if (plink->d8 == NULL)
	{
          plink->d8 = links + loffs;
	  plink = plink->d8;
	  plink->d4 = NULL;
	  plink->d8 = NULL;
	  loffs++;
	}
	else
	  plink = plink->d8;
      }
      if (dn != 1)
	plink->d0 = 257;
      else
	plink->d0 = cnt;
      dn--;
    }//while (dn != 0)
    cnt++;
  }//while (cnt <= 256)
}
/*decode()*/

//-------------------------------------------------------
//vim: syntax=cpp

